import io.github.jaredmdobson.concentus.OpusApplication
import io.github.jaredmdobson.concentus.OpusDecoder
import io.github.jaredmdobson.concentus.OpusEncoder
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import java.io.IOException
import java.io.ObjectInputStream
import java.io.ObjectOutputStream
import java.net.ServerSocket
import java.net.Socket
import java.util.*
import java.util.concurrent.CopyOnWriteArrayList

val clientAudios = CopyOnWriteArrayList<ClientAudio>()

object Server {

    @JvmStatic
    fun main(args: Array<String>) {
        mainServer()
    }

    fun mainServer() {

        val scope = CoroutineScope(Dispatchers.IO)
        scope.launch {
            while (true) {
                for (client in clientAudios) {
                    if (!client.isConnected) {
                        clientAudios -= client
                        continue
                    }
                    client.read()
                }

                for (client in clientAudios) {
                    launch {
                        for (audio in clientAudios) {
                            if (audio != client || !ECHO_CANCELLATION) {
                                client.mix(audio.sample)
                            }
                        }
                        client.send()
                    }
                }
            }
        }

        val sever = ServerSocket(SERVER_PORT)
        while (true) {
            val socket = sever.accept()
            println("Accepted connection from " + socket.remoteSocketAddress)
            clientAudios += ClientAudio(socket)
        }
    }

}

class ClientAudio(val socket: Socket) {

    val inp = ObjectInputStream(socket.getInputStream())
    val out = ObjectOutputStream(socket.getOutputStream())
    val buffer = ByteArray(BUFFER_SIZE)
    val sample = FloatArray(PACKET_SAMPLES)
    val mixSample = FloatArray(PACKET_SAMPLES)
    val mixBuffer = ByteArray(BUFFER_SIZE)

    val decoder = OpusDecoder(SAMPLE_RATE, CHANNELS)
    val encoder = OpusEncoder(SAMPLE_RATE, CHANNELS, OpusApplication.OPUS_APPLICATION_AUDIO).also {
        it.bitrate = OPUS_BITRATE
        it.forceMode = OPUS_MODE
        it.signalType = OPUS_SIGNAL_TYPE
        it.complexity = OPUS_COMPLEXITY
    }

    var isConnected = true

    fun send() {
        try {
            AudioConvert.convertFloatToShortByte(mixSample, mixBuffer)
            out.writeObject(encoder.encode(mixBuffer))
            out.flush()
        } catch (e: IOException) {
            isConnected = false
        }
    }

    fun read() {
        try {
//            if (`in`!!.available() > bufferIn.size * 2) {
//                `in`!!.skip(`in`!!.available().toLong() - bufferIn.size)
//                println("丢弃")
//            }
//            `in`.read(buffer)
            val pack = inp.readObject() as DataPack
//            println("延迟： ${System.currentTimeMillis() - timestamp}ms")
            decoder.decode(pack, buffer)
            AudioConvert.convertShortByteToFloat(buffer, sample)
            Arrays.fill(mixSample, 0f)
        } catch (e: IOException) {
            isConnected = false
        }
    }

    fun mix(sample: FloatArray) {
        for (i in sample.indices) {
            mixSample[i] += sample[i]
        }
    }
}
